#include <math.h>
#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <sys/time.h>
#include <unistd.h>
#include <sys/resource.h>
#include "ASL_user.h"
#include "tbb/blocked_range.h"
#include "tbb/parallel_for.h"
#include "tbb/task_scheduler_init.h"
using namespace tbb ;

#define FALSE 0
#define TRUE 1
#define MAX(a,b) (((a) > (b)) ? (a) : (b))
#define MIN(a,b) (((a) < (b)) ? (a) : (b))
#define ZERO ((double) 0)

double Timer ( void ) ;

double WallTimer ( void ) ;

typedef struct SVMproblem_struct /* passed to user's evaluation routines */
{
    float      *Ax ; /* numerical values in matrix */
    short      *Ai ; /* row indices */
    int        *Ap ; /* pointers to start of each column in Ax, Ai */
    short       *a ; /* label vector (+- 1) */
    float       *A ; /* dense A matrix */
    int        *AP ; /* pointers to start of each column in A */
    int       *col ; /* list of columns to build */
    int          n ; /* problem dimension */
    double  lo_cut ; /* x_i < lo_cut => x_i at lower bound */
    double   *temp ; /* temporary storage when computing A times vector */
    int   nfeature ; /* number of features */
    short Ax_is_one ; /* TRUE = all elements of Ax are 1 */
    short   kernel ; /* 0 = linear, 1 = poly, 2 = rbf, 3 = sig */
    short   degree ; /* parameter in kernel */
    float    gamma ; /* parameter in kernel */
    int        mem ; /* location of free space in A */
    double val_time ; /* time spent evaluating columns (nonlinear) */
    double  Ad_time ; /* time spent evaluating Ad (including columns) */
    int      nproc ; /* number of processors */
    double  *pwork ; /* parallel workspace, size n*nproc */
    double  *rwork ; /* parallel workspace, size n*nproc */
    int    *bstart ; /* start of column blocks for pwork */
    ASLobjective *data ; /* pointer to objective structure */
} SVMproblem ;

void read_problem
(
    const char *filename
) ;

float sparse_dot
(
    float  *x,  /* nonzero numerical values */
    short *xi,  /* nonzero indices in x */
    int   nzx,  /* number of nonzeros in x */
    float  *y,  /* nonzero numerical values */
    short *yi,  /* nonzero indices in y */
    int   nzy   /* number of nonzeros in y */
) ;

float binary_sparse_dot
(
    short *xi,  /* nonzero indices in x */
    int   nzx,  /* number of nonzeros in x */
    short *yi,  /* nonzero indices in y */
    int   nzy   /* number of nonzeros in y */
) ;

float sparse_norm2
(
    float  *x,  /* nonzero numerical values */
    short *xi,  /* nonzero indices in x */
    int   nzx,  /* number of nonzeros in x */
    float  *y,  /* nonzero numerical values */
    short *yi,  /* nonzero indices in y */
    int   nzy   /* number of nonzeros in y */
) ;

float binary_sparse_norm2
(
    short *xi,  /* nonzero indices in x */
    int   nzx,  /* number of nonzeros in x */
    short *yi,  /* nonzero indices in y */
    int   nzy   /* number of nonzeros in y */
) ;

void scalar_init
(
    double  *x,  /* numerical values */
    double   s,  /* scalar */
    int      n   /* length of vectors */
) ;

void int_init
(
    int *x,  /* numerical values */
    int  s,  /* scalar */
    int  n   /* length of vectors */
) ;

void dense_daxpy
(
    double    *x,
    float     *y,
    double     s, /* scalar */
    int        n  /* dimension */
) ;

void dense_scalar_mult
(
    double    *x,
    float     *y,
    double     s, /* scalar */
    int        n  /* dimension */
) ;

void row_add
(
    double    *x,
    double    *y,
    int        n  /* dimension */
) ;

void get_col
(
    int k   /* build k-th column of A */
) ;

void build_col_work
(
    int j   /* build column j of pwork, store in rwork (row version of pwork) */
) ;

void build_rows
(
    int k   /* build row block k of d */
) ;

void build_p
(
    int k   /* build row block k of d */
) ;

void SVMlinearAd
(
    ASLobjective *data
) ;

void SVMnonlinearAd
(
    ASLobjective *data
) ;

class BuildMatrix
{
    public:
        void operator ()(const blocked_range <int> & r) const {
            int i ;
            for ( i = r.begin(); i!= r.end(); i++ )
            {
                get_col (i) ;

            }
        }
        BuildMatrix ( )
        {}
} ;

class BuildPwork
{
    public:
        void operator ()(const blocked_range <int> & r) const {
            int i ;
            for ( i = r.begin(); i!= r.end(); i++ )
            {
                build_col_work (i) ;

            }
        }
        BuildPwork ( )
        {}
} ;

class BuildD
{
    public:
        void operator ()(const blocked_range <int> & r) const {
            int i ;
            for ( i = r.begin(); i!= r.end(); i++ )
            {
                build_rows (i) ;

            }
        }
        BuildD ( )
        {}
} ;

class BuildP
{
    public:
        void operator ()(const blocked_range <int> & r) const {
            int i ;
            for ( i = r.begin(); i!= r.end(); i++ )
            {
                build_p (i) ;

            }
        }
        BuildP ( )
        {}
} ;

void SequentialBuildMatrix
(
    int ncol   /* number of columns to build */
) ;

void SequentialBuildPwork
(
    int nproc  /* number of processors */
) ;

void SequentialBuildD
(
    int nproc  /* number of processors */
) ;

void SequentialBuildP
(
    int nproc  /* number of processors */
) ;

void ParallelBuildMatrix (int n, int GrainSize)
{
    parallel_for (blocked_range <int>(0, n, GrainSize), BuildMatrix ( ) ) ;
}

void ParallelBuildPwork (int n, int GrainSize)
{
    parallel_for (blocked_range <int>(0, n, GrainSize), BuildPwork ( ) ) ;
}

void ParallelBuildD (int n, int GrainSize)
{
    parallel_for (blocked_range <int>(0, n, GrainSize), BuildD ( ) ) ;
}

void ParallelBuildP (int n, int GrainSize)
{
    parallel_for (blocked_range <int>(0, n, GrainSize), BuildP ( ) ) ;
}

SVMproblem PB ;

int main (int argc, char **argv)
{
    task_scheduler_init init ;
    short kernel, *Ai ;
    int i, j, k, n, nfeature, Apn, nlo, nhi, nfree, ncol, *Ap ;
    long N ;
    float *A, *Ax ;
    double b, C, lo, hi, t, tic, tic0, tol, xj, atx, *x, *g0, *temp ;
    ASLstat Stat ;
    ASLparm Parm ;
    C  = atof (argv [3]) ;
    printf ("\n\nproblem: %s // kernel: %s // C: %e\n",
             argv [1]+9, argv [2], C) ;
    PB.nproc = 15 ;
    tic = WallTimer () ;
    tic0 = tic ;
    read_problem (argv [1]) ;
    printf ("time to read data: %e\n", WallTimer () - tic) ;
    n = PB.n ;
    g0 = (double *) malloc (n*sizeof (double)) ;
    x = (double *) malloc (n*sizeof (double)) ;
    /* convert problem to standard form with a_i = 1 for all i
       x_i = -y_i + C if a_i = -1
       starting guess:    x_i = 0 for all i or
                          y_i = C if a_i = -1
                          x_i = y_i = 0 otherwise
       starting gradient: g_i = +1 if a_i = -1
                          g_i = -1 otherwise */
    b = 0. ;
    for (j = 0; j < n; j++)
    {
        if ( PB.a [j] < 0 )
        {
            x [j] = C ;
            b += C ;
            g0 [j] = +1 ;
        }
        else
        {
            x [j] = 0 ;
            g0 [j] = -1 ;
        }
    }
    PB.val_time = 0. ;
    PB.Ad_time = 0. ;
    N = (long) n ;
    nfeature = PB.nfeature ;
    Ap = PB.Ap ;
    Ai = PB.Ai ;
    Ax = PB.Ax ;
    Apn = Ap [n] ;
    printf ("number of features: %i\n", PB.nfeature) ;
    printf ("nonzeros in data: %i\n", Apn) ;
    printf ("problem dimension: %i\n", n) ;
    fflush (stdout) ;
 
    printf ("matrix to be created in format %s\n", argv [2]) ;
    fflush (stdout) ;
    kernel = 0 ;
    if ( strcmp (argv [2],"linear") ) /* if not linear, compute A */
    {
        if      ( !strcmp (argv [2], "poly") ) kernel = 1 ;
        else if ( !strcmp (argv [2], "rbf") )  kernel = 2 ;
        else if ( !strcmp (argv [2], "sig") )  kernel = 3 ;
        else
        {
            printf ("%s not a valid kernel\n", argv [2]) ;
            exit (0) ;
        }
    }
    PB.kernel = kernel ;

    /* partition columns or rows among processors */
    PB.bstart = (int *) malloc ((PB.nproc +1)*sizeof (int)) ;
    if ( !strcmp (argv [2],"linear") )
    {
        temp = PB.temp = (double *) malloc (nfeature*sizeof (double)) ;
    }
    else
    {
        A = PB.A = (float *) malloc (N*N*sizeof (float)) ;
        PB.AP = (int *) malloc (N*sizeof (int)) ; /*pointer to location of col*/
        int_init (PB.AP, -1, n) ;
        PB.col = (int *) malloc (N*sizeof (int)) ; /* columns to build */
        PB.pwork = (double *) malloc (PB.nproc*N*sizeof (double)) ;
        PB.rwork = (double *) malloc (PB.nproc*N*sizeof (double)) ;
        PB.degree = 3 ;
        PB.gamma = 1/ (float) nfeature ;
        PB.mem = 0 ;                              /* free memory */
    }

    /* setup for ASL */
    lo = 0. ;
    hi = C ;
    tol = 1.e-3 ;
/*  if ( C > 9 ) tol = 1.e-2 ;*/
    ASLdefault (&Parm) ;

/*  Parm.maxits = 700 ;*/
    Parm.PrintLevel = 0 ;
    Parm.lo_cut = 1.e-12 ;
    Parm.hi_cut = 1.e-12 ;
    Parm.QuadCost = TRUE ;
    Parm.a_is_1 = TRUE ;
    Parm.QuadLazy = FALSE ;
    Parm.bnd_is_scalar = TRUE ;
    if ( !strcmp (argv [2],"linear") )
    {
       Parm.Subspace = FALSE ;
    }
    else
    {
        Parm.Subspace = TRUE ;
        Parm.nsub = 250 ;
    }

    /* usually tol = 1.e-3 */
    tic = WallTimer () ;
    if ( !strcmp (argv [2],"linear") )
    {
        printf ("call linear solver\n") ;
        fflush (stdout) ;
        ASL (x, &lo, &hi, NULL, b, g0, n, tol, &Stat, &Parm,
             SVMlinearAd, NULL, NULL, NULL) ;
    }
    else
    {
        printf ("call nonlinear solver\n") ;
        fflush (stdout) ;
        ASL (x, &lo, &hi, NULL, b, g0, n, tol, &Stat, &Parm,
             SVMnonlinearAd, NULL, NULL, NULL) ;
    }
    printf ("ASL time: %e\n", WallTimer () - tic) ;
    printf ("Total time (read+matrix+ASL): %e\n", WallTimer () - tic0) ;
    printf ("Time in evaluation of cols of A: %e\n", PB.val_time) ;
    printf ("Time in evaluation of Ad: %e\n", PB.Ad_time) ;
    printf ("Function value from stat: %e\n", Stat.f) ;
    ncol = 0 ;
    nfree = 0 ;
    nhi = 0 ;
    nlo = 0 ;
    atx = 0. ;
    /* count bound and free components of x */
    for (j = 0; j < n; j++)
    {
        /* reverse the transformation on x */
        if ( PB.a [j] < 0 ) x [j] = C - x [j] ;
        if      ( x [j] < 1.e-6 )     nlo++ ;
        else if ( x [j] > C - 1.e-6 ) nhi++ ;
        else                          nfree++ ;
        atx += PB.a [j]*x [j] ;
        /* count number of evaluated columns if problem nonlinear */
        if ( strcmp (argv [2],"linear") ) if ( PB.AP [j] >= 0 ) ncol++ ;
    }
    /* compute objective function value at x */
    scalar_init (g0, 0, n) ;
    if ( !strcmp (argv [2],"linear") ) scalar_init (temp, 0, nfeature) ;
    for (j = 0; j < n; j++)
    {
        if ( x [j] != ZERO )
        {
            xj = .5*x [j] ;
            if ( PB.a [j] < 0. ) xj = -xj ; /* a [j] enters the columns */
            if ( !strcmp (argv [2],"linear") )
            {
                for (k = Ap [j]; k < Ap [j+1]; k++)
                {
                    temp [Ai [k]] += xj*Ax [k] ;
                }
            }
            else
            {
                if ( PB.AP [j] >= 0 ) dense_daxpy (g0, A+(PB.AP [j]*N), xj, n) ;
                else printf ("j: %i xj: %e\n", j, 2*fabs(xj)) ;
            }
        }
    }
    if ( !strcmp (argv [2],"linear") ) /* complete the evaluation of Ax */
    {
        for (j = 0; j < n; j++)
        {
            t = 0 ;
            for (k = Ap [j]; k < Ap [j+1]; k++) t += Ax [k]*temp [Ai [k]] ;
            g0 [j] = t ;
        }
    }
    printf ("nlo: %i nhi: %i nfree: %i\n", nlo, nhi, nfree) ;
    printf ("linear constraint violation: %e\n", atx) ;
    printf ("number of evaluated columns: %i\n", ncol) ;
    /* account for linear term and -1s in quadratic part of cost */
    t = ZERO ;
    for (j = 0; j < n; j++)
    {
        if ( PB.a [j] < 0 ) t += (-1 - g0 [j])*x [j] ;
        else                t += (-1 + g0 [j])*x [j] ;
        
    }
    printf ("FINAL COST: %e\n", t) ;
    free (PB.Ax) ;
    free (PB.Ai) ;
    free (PB.Ap) ;
    free (PB.AP) ;
    free (PB.a) ;
    free (x) ;
    free (g0) ;
    free (PB.bstart) ;
    if ( !strcmp (argv [2],"linear") )
    {
        free (PB.temp) ;
    }
    else
    {
        free (PB.A) ;
        free (PB.col) ;
        free (PB.pwork) ;
        free (PB.rwork) ;
    }
    i = argc ;
    return 0 ;
    return (1) ;
}

/* sparse dot product x'y */
float sparse_dot
(
    float  *x,  /* nonzero numerical values */
    short *xi,  /* nonzero indices in x */
    int   nzx,  /* number of nonzeros in x */
    float  *y,  /* nonzero numerical values */
    short *yi,  /* nonzero indices in y */
    int   nzy   /* number of nonzeros in y */
)
{
    int i, j ;
    float t ;
    t = 0. ;
    i = 0 ;
    j = 0 ;
    while ( (i < nzx) && (j < nzy) )
    {
        if ( yi [j] == xi [i] )
        {
            t += y [j]*x [i] ;
            i++ ;
            j++ ;
        }
        else
        {
            if ( yi [j] > xi [i] ) i++ ;
            else                   j++ ;
        }
    }
    return (t) ;
}

/* binary sparse dot product x'y (entries in x and y are 0 or 1) */
float binary_sparse_dot
(
    short *xi,  /* nonzero indices in x */
    int   nzx,  /* number of nonzeros in x */
    short *yi,  /* nonzero indices in y */
    int   nzy   /* number of nonzeros in y */
)
{
    int i, j ;
    float t ;
    t = 0. ;
    i = 0 ;
    j = 0 ;
    while ( (i < nzx) && (j < nzy) )
    {
        if ( yi [j] == xi [i] )
        {
            t++ ;
            i++ ;
            j++ ;
        }
        else
        {
            if ( yi [j] > xi [i] ) i++ ;
            else                   j++ ;
        }
    }
    return (t) ;
}

/* ||x_i - x_j||^2 */
float sparse_norm2
(
    float  *x,  /* nonzero numerical values */
    short *xi,  /* nonzero indices in x */
    int   nzx,  /* number of nonzeros in x */
    float  *y,  /* nonzero numerical values */
    short *yi,  /* nonzero indices in y */
    int   nzy   /* number of nonzeros in y */
)
{
    int i, j ;
    float s, t ;
    s = 0. ;
    i = 0 ;
    j = 0 ;
    while ( (i < nzx) && (j < nzy) )
    {
        if ( yi [j] == xi [i] )
        {
            t = y [j] - x [j] ;
            s += t*t ;
            i++ ;
            j++ ;
        }
        else
        {
            if ( yi [j] > xi [i] )
            {
                t = x [i] ;
                s += t*t ;
                i++ ;
            }
            else
            {
                t = y [j] ;
                s += t*t ;
                j++ ;
            }
        }
    }
    if ( i < nzx )
    {
        for (; i < nzx; i++)
        {
            t = x [i] ;
            s += t*t ;
        }
    }
    else
    {
        while ( j < nzy )
        {
            t = y [j] ;
            s += t*t ;
            j++ ;
        }
    }
    return (s) ;
}
/* ||x_i - x_j||^2 with x_i and x_j  binary */
float binary_sparse_norm2
(
    short *xi,  /* nonzero indices in x */
    int   nzx,  /* number of nonzeros in x */
    short *yi,  /* nonzero indices in y */
    int   nzy   /* number of nonzeros in y */
)
{
    int i, j ;
    float s ;
    s = 0. ;
    i = 0 ;
    j = 0 ;
    while ( (i < nzx) && (j < nzy) )
    {
        if ( yi [j] == xi [i] )
        {
            i++ ;
            j++ ;
        }
        else
        {
            s++ ;
            if ( yi [j] > xi [i] ) i++ ;
            else                   j++ ;
        }
    }
    s += (float) ((nzx-i) + (nzy-j)) ;
    return (s) ;
}

void scalar_init
(
    double  *x,  /* numerical values */
    double   s,  /* scalar */
    int      n   /* length of vectors */
)
{
    int j, n5 ;
    double *xj ;
    n5 = n % 5 ;
    for (j = 0; j < n5; j++) x [j] = s ;
    xj = x+j ;
    for (; j < n; j += 5)
    {
        *(xj++) = s ;
        *(xj++) = s ;
        *(xj++) = s ;
        *(xj++) = s ;
        *(xj++) = s ;
    }
}

void int_init
(
    int *x,  /* numerical values */
    int  s,  /* scalar */
    int  n   /* length of vectors */
)
{
    int j, n5 ;
    int *xj ;
    n5 = n % 5 ;
    for (j = 0; j < n5; j++) x [j] = s ;
    xj = x+j ;
    for (; j < n; j += 5)
    {
        *(xj++) = s ;
        *(xj++) = s ;
        *(xj++) = s ;
        *(xj++) = s ;
        *(xj++) = s ;
    }
}

/* =========================================================================
   === dense_daxpy =========================================================
   =========================================================================
       set x = x + s*y
   ========================================================================= */
void dense_daxpy
(
    double    *x,
    float     *y,
    double     s, /* scalar */
    int        n  /* dimension */
)
{
    int j, n5 ;
    n5 = n % 5 ;
    for (j = 0; j < n5; j++) x [j] += s*y [j] ;
    for (; j < n; j += 5)
    {
        x [j  ] += s*y [j  ] ;
        x [j+1] += s*y [j+1] ;
        x [j+2] += s*y [j+2] ;
        x [j+3] += s*y [j+3] ;
        x [j+4] += s*y [j+4] ;
    }
}

/* =========================================================================
   === dense_scalar_mult ===================================================
   =========================================================================
       set x = s*y
   ========================================================================= */
void dense_scalar_mult
(
    double    *x,
    float     *y,
    double     s, /* scalar */
    int        n  /* dimension */
)
{
    int j, n5 ;
    n5 = n % 5 ;
    for (j = 0; j < n5; j++) x [j] = s*y [j] ;
    for (; j < n; j += 5)
    {
        x [j  ] = s*y [j  ] ;
        x [j+1] = s*y [j+1] ;
        x [j+2] = s*y [j+2] ;
        x [j+3] = s*y [j+3] ;
        x [j+4] = s*y [j+4] ;
    }
}

/* =========================================================================
   === row_add =============================================================
   =========================================================================
       add entries in a row and store in x, assuming n <= 16
   ========================================================================= */
void row_add
(
    double    *x,
    double    *y,
    int        n  /* dimension */
)
{
    int left ;
    double t, *yp ;
    if ( n == 16 )
    {
        *x = y[0]+y[1]+y[2]+y[3]+y[4]+y[5]+y[6]+y[7]+y[8]+y[9]+y[10]+
                  y[11]+y[12]+y[13]+y[14]+y[15] ;
        return ;
    }
    else if ( n >= 12 )
    {
        t = y[0]+y[1]+y[2]+y[3]+y[4]+y[5]+y[6]+y[7]+y[8]+y[9]+y[10]+y[11] ;
        yp = y+12 ;
        left = n - 12 ;
    }
    else if ( n >= 8 )
    {
        t = y[0]+y[1]+y[2]+y[3]+y[4]+y[5]+y[6]+y[7] ;
        yp = y+8 ;
        left = n - 8 ;
    }
    else if ( n >= 4 )
    {
        t = y[0]+y[1]+y[2]+y[3] ;
        yp = y+4 ;
        left = n - 4 ;
    }
    else { t = 0. ; left = n ; yp = y ; }
    if      ( left == 3 ) *x = t + yp[0] + yp[1] + yp[2] ;
    else if ( left == 2 ) *x = t + yp[0] + yp[1] ;
    else if ( left == 1 ) *x = t + yp[0] ;
    else                  *x = t ;
}
    
void get_col
(
    int j   /* build column PB.cols [k] of A */
)
{
    int k, n, nzx, nzy, *Ap ;
    short degree, *Ai, *Aij ;
    float gamma, t, *A, *Ax, *Axj ;
    long K, N ;
    n = PB.n ;
    N = (long) n ;
    k = PB.col [j] ;
    K = N*PB.AP [k] ;
    Ap = PB.Ap ;
    Ai = PB.Ai ;
    Ax = PB.Ax ;
    A = PB.A ;
    if ( PB.kernel == 1 ) /* poly */
    {
        gamma = PB.gamma ;
        degree = PB.degree ;
        if ( PB.Ax_is_one )
        {
            nzx = Ap [k+1] - Ap [k] ;
            Aij = Ai+Ap [k] ;
            for (j = 0; j < n; j++)
            {
                nzy = Ap [j+1] - Ap [j] ;
                t = binary_sparse_dot (Aij, nzx, Ai+Ap[j], nzy) ;
                t = gamma*t ;
                t = t*t*t ;
                A [K] = t ;
                K++ ;
            }
        }
        else
        {
            nzx = Ap [k+1] - Ap [k] ;
            Axj = Ax+Ap [k] ;
            Aij = Ai+Ap [k] ;
            for (j = 0; j < n; j++)
            {
                nzy = Ap [j+1] - Ap [j] ;
                t = sparse_dot (Axj, Aij, nzx, Ax+Ap[j], Ai+Ap[j], nzy);
                t = gamma*t ;
                t = t*t*t ;
                A [K] = t ;
                K++ ;
            }
        }
    }
    else if ( PB.kernel == 2 ) /* rbf */
    {
        gamma = PB.gamma ;
        if ( PB.Ax_is_one )
        {
            nzx = Ap [k+1] - Ap [k] ;
            Aij = Ai+Ap [k] ;
            for (j = 0; j < n; j++)
            {
                nzy = Ap [j+1] - Ap [j] ;
                t = binary_sparse_norm2 (Aij, nzx, Ai+Ap[j], nzy) ;
                t = exp (-gamma*t) ;
                A [K] = t ;
                K++ ;
            }
        }
        else
        {
            nzx = Ap [k+1] - Ap [k] ;
            Axj = Ax+Ap [k] ;
            Aij = Ai+Ap [k] ;
            for (j = 0; j < n; j++)
            {
                nzy = Ap [j+1] - Ap [j] ;
                t = sparse_norm2 (Axj, Aij, nzx, Ax+Ap[j],Ai+Ap[j],nzy);
                t = exp (-gamma*t) ;
                A [K] = t ;
                K++ ;
            }
        }
    }
    else                       /* sig */
    {
        gamma = PB.gamma ;
        if ( PB.Ax_is_one )
        {
            nzx = Ap [k+1] - Ap [k] ;
            Aij = Ai+Ap [k] ;
            for (j = 0; j < n; j++)
            {
                nzy = Ap [j+1] - Ap [j] ;
                t = binary_sparse_dot (Aij, nzx, Ai+Ap[j], nzy) ;
                t = tanh (gamma*t) ;
                A [K] = t ;
                K++ ;
            }
        }
        else
        {
            nzx = Ap [k+1] - Ap [k] ;
            Axj = Ax+Ap [k] ;
            Aij = Ai+Ap [k] ;
            for (j = 0; j < n; j++)
            {
                nzy = Ap [j+1] - Ap [j] ;
                t = sparse_dot (Axj, Aij, nzx, Ax+Ap[j], Ai+Ap[j], nzy);
                t = tanh (gamma*t) ;
                A [K] = t ;
                K++ ;
            }
        }
    }
}

void build_col_work
(
    int j   /* build column j of pwork, store in rwork (row version of pwork) */
)
{
    int k, l, n, n5, nproc, ncols, *bstart, *i ;
    double *col, *colk, *d, *row ;
    long N ;
    n = PB.n ;
    N = n ;
    bstart = PB.bstart ;
    k = bstart [j] ;
    l = bstart [j+1] ;
    if ( l == k ) return ;
    col = PB.pwork+(N*j) ; /* first generate column in pwork */
    i = PB.data->i ;
    d = PB.data->d ;
    nproc = PB.nproc ;
    dense_scalar_mult (col, PB.A+(PB.AP [i [k]]*N), d [k], n) ;
    k++ ;
    for (; k < l; k++)
    {
        dense_daxpy (col, PB.A+(PB.AP [i [k]]*N), d [k], n) ;
    }
    /* copy column into rwork, where the pwork array is stored by rows */
    row = PB.rwork+j ;
    ncols = MIN (nproc, PB.data->nz) ;
    n5 = n % 5 ;
    for (k = 0; k < n5; k++)
    {
        *row = col [k] ;
         row = row+ncols ;
    }
    colk = col+k ;
    for (; k < n; k += 5)
    {
        *row = *(colk++) ;
         row = row+ncols ;
        *row = *(colk++) ;
         row = row+ncols ;
        *row = *(colk++) ;
         row = row+ncols ;
        *row = *(colk++) ;
         row = row+ncols ;
        *row = *(colk++) ;
         row = row+ncols ;
    }
}

void build_p
(
    int k   /* build row block k of product vector p */
)
{
    int j, jend, l, m, *bstart, *Ap, *Ap1 ;
    double t, *p, *temp ;
    float *Ax ;
    short *Ai ;
    p = PB.data->p ;
    temp = PB.temp ;
    bstart = PB.bstart ;
    Ai = PB.Ai ;
    Ap = PB.Ap ;
    Ap1= Ap+1 ;
    Ax = PB.Ax ;
    j = bstart [k] ;
    jend = bstart [k+1] ;
    if ( j == jend ) return ;

    if ( PB.Ax_is_one )
    {
        m = Ap [j] ;
        for (; j < jend; j++)
        {
            t = ZERO ;
            l = Ap1 [j] ;
            for (; m < l; m++) t += temp [Ai [m]] ;
            p [j] = t ;
        }
    }
    else
    {
        m = Ap [j] ;
        for (; j < jend; j++)
        {
            t = ZERO ;
            l = Ap1 [j] ;
            for (; m < l; m++) t += temp [Ai [m]]*Ax [m] ;
            p [j] = t ;
        }
    }
}

void build_rows
(
    int k   /* build row block k of d */
)
{
    int i, l, nproc, ncols, *bstart ;
    double *p, *row ;
    p = PB.data->p ;
    bstart = PB.bstart ;
    i = bstart [k] ;
    l = bstart [k+1] ;
    if ( i == l ) return ;
    nproc = PB.nproc ;
    ncols = MIN (nproc, PB.data->nz) ;
    row = PB.rwork+(bstart [k]*ncols) ;
    for (; i < l; i++)
    {
        row_add (p+i, row, ncols) ;
        row = row+ncols ;
    }
}

void SequentialBuildMatrix
(
    int ncol   /* number of columns to build */
)
{
    int k ;
    for (k = 0; k < ncol; k++) get_col (k) ;
}

void SequentialBuildPwork
(
    int nproc  /* number of processors */
)
{
    int k ;
    for (k = 0; k < nproc; k++) build_col_work (k) ;
}

void SequentialBuildD
(
    int nproc  /* number of processors */
)
{
    int k ;
    for (k = 0; k < nproc; k++) build_rows (k) ;
}

void SequentialBuildP
(
    int nproc  /* number of processors */
)
{
    int k ;
    for (k = 0; k < nproc; k++) build_p (k) ;
}

/* SVM nonlinear Ad */
void SVMnonlinearAd
(
    ASLobjective *data
)
{
    int bsize, bleft, j, k, nz, n, ncol, nproc, grainsize, *bstart, *i, *AP ;
    double tic, tic0, *d, *p ;
    float *A ;
    long N ;
    tic0 = WallTimer () ;
    bstart = PB.bstart ;
    nproc = PB.nproc ;
    PB.data = data ;
    nz = data->nz ;
    i = data->i ;
    d = data->d ;
    p = data->p ;
    AP = PB.AP ;
    A = PB.A ;
    n = PB.n ;
    N = n ;
    /* determine the columns that we need to build */
    ncol = 0 ;
    for (k = 0; k < nz; k++)
    {
        j = i [k] ;
        if ( AP [j] < 0 )
        {
            PB.col [ncol] = j ;
            PB.AP [j] = PB.mem ;
            PB.mem++ ;
            ncol++ ;
        }
    }
    grainsize = ncol/nproc ;
    if ( grainsize*nproc < ncol ) grainsize++ ;
    tic = WallTimer () ;
    if ( ncol ) ParallelBuildMatrix (ncol, grainsize) ;
/*  if ( ncol ) SequentialBuildMatrix (ncol) ;*/
    PB.val_time += WallTimer () - tic ;
    /* partition columns amoung processors */
    bsize = nz/nproc ;
    bleft = nz - bsize*nproc ;
    bstart [0] = 0 ;
    for (k = 0; k < bleft; k++)
    {
        bstart [k+1] = bstart [k] + bsize + 1 ;
    }
    for (; k < nproc; k++)
    {
        bstart [k+1] = bstart [k] + bsize ;
    }
    ParallelBuildPwork (nproc, 1) ;
/*  SequentialBuildPwork (nproc) ;*/
    /* partition the rows of rwork among the processors */
    bsize = n/nproc ;
    bleft = n - bsize*nproc ;
    for (k = 0; k < bleft; k++)
    {
        bstart [k+1] = bstart [k] + bsize + 1 ;
    }
    for (; k < nproc; k++)
    {
        bstart [k+1] = bstart [k] + bsize ;
    }
    ParallelBuildD (nproc, 1) ;
/*  SequentialBuildD (nproc) ;*/
    PB.Ad_time += WallTimer () - tic0 ;
}

/* SVM linear Ad */
void SVMlinearAd
(
    ASLobjective *data
)
{
    int bsize, bleft, nproc, j, k, l, m, n, nz, *i, *Ap, *Ap1, *bstart ;
    double dk, t, tic, *d, *p, *temp ;
    float *Ax ;
    short *Ai ;
    tic = WallTimer () ;
    PB.data = data ;
    n = PB.n ;
    Ai = PB.Ai ;
    Ap = PB.Ap ;
    Ap1= Ap+1 ;
    Ax = PB.Ax ;
    temp = PB.temp ;
    nproc = PB.nproc ;
    bstart = PB.bstart ;
    nz = data->nz ;
    i = data->i ;
    d = data->d ;
    p = data->p ;
    scalar_init (temp, 0., PB.nfeature) ;

    /* partition the rows of rwork among the processors */
    bsize = n/nproc ;
    bleft = n - bsize*nproc ;
    bstart [0] = 0 ;
    for (k = 0; k < bleft; k++)
    {
        bstart [k+1] = bstart [k] + bsize + 1 ;
    }
    for (; k < nproc; k++)
    {
        bstart [k+1] = bstart [k] + bsize ;
    }

    if ( PB.Ax_is_one )
    {
        for (m = 0; m < nz; m++)
        {
            dk = d [m] ;
            j = i [m] ;
            l = Ap1 [j] ;
            for (k = Ap [j]; k < l; k++) temp [Ai [k]] += dk ;
        }
/*      k = 0 ;
        for (j = 0; j < n; j++)
        {
            t = ZERO ;
            l = Ap1 [j] ;
            for (; k < l; k++) t += temp [Ai [k]] ;
            p [j] = t ;
        } */
        ParallelBuildP (nproc, 1) ;
/*      SequentialBuildP (nproc) ;*/
    }
    else
    {
        for (m = 0; m < nz; m++)
        {
            dk = d [m] ;
            j = i [m] ;
            l = Ap1 [j] ;
            for (k = Ap [j]; k < l; k++) temp [Ai [k]] += dk*Ax [k] ;
        }
/*      k = 0 ;
        for (j = 0; j < n; j++)
        {
            t = ZERO ;
            l = Ap1 [j] ;
            for (; k < l; k++) t += temp [Ai [k]]*Ax [k] ;
            p [j] = t ;
        } */
        ParallelBuildP (nproc, 1) ;
/*      SequentialBuildP (nproc) ;*/
    }
    PB.Ad_time += WallTimer () - tic ;
}

/* read in a problem (in svmlight format) */
void read_problem
(
    const char *filename
)
{
    int elements, i, j, k, max_index, n ;
    float *Ax, Aj ;
    int *Ap ;
    short *a, *Ai, Ax_is_one ;
    FILE *fp = fopen(filename,"r");

    if(fp == NULL)
    {
        fprintf(stderr,"can't open input file %s\n",filename);
        exit(1);
    }

    n = 0 ;
    elements = 0 ;
    while(1)
    {
        int c = fgetc(fp) ;
        switch(c)
        {
            case '\n':
                ++n ; break ; 
            case ':':
                ++elements ; break;
            case EOF:
                goto out;
            default:
            ;
        }
    }
out:
    rewind(fp);

    PB.a = a = (short *) malloc (n*sizeof (short));
    PB.Ax = Ax = (float  *) malloc (elements*sizeof (float)) ;
    PB.Ai = Ai = (short *) malloc (elements*sizeof (short)) ;
    PB.Ap = Ap = (int *) malloc ((n+1)*sizeof (int)) ;
    PB.n = n ;

    Ap [0] = 0 ;
    max_index = 0 ;
    Ax_is_one = TRUE ;
    j = 0 ;
    for(i = 0; i < n; i++)
    {
        fscanf(fp,"%d", &k);
        a [i] = k ;

        while(1)
        {
            int c;
            do {
                c = getc(fp) ;
                if ( c=='\n' ) goto out2 ;
            } while (isspace(c)) ;
            ungetc(c,fp);
            if ( fscanf(fp,"%d:%f", &k, &Aj) < 2 )
            {
              fprintf(stderr,"Wrong input format at line %d\n", i+1) ;
              exit(1) ;
            }
            Ai [j] = k-1 ;
            if ( Aj != 1 ) Ax_is_one = FALSE ;
            Ax [j] = Aj ;
            j++ ;
        }
out2:
        if ( (j >= 1) && k > max_index ) max_index = k ;
        Ap [i+1] = j ;
    }
    PB.nfeature = max_index ;
    PB.Ax_is_one = Ax_is_one ;
    fclose(fp);
}

double Timer ( void )
{
    struct rusage ru ;
    double user, tsys ;

    (void) getrusage (RUSAGE_SELF, &ru) ;

    user =
    ru.ru_utime.tv_sec                  /* user time (seconds) */
    + 1e-6 * ru.ru_utime.tv_usec ;      /* user time (microseconds) */

    tsys =
    ru.ru_stime.tv_sec                  /* tsys time (seconds) */
    + 1e-6 * ru.ru_stime.tv_usec ;      /* tsys time (microseconds) */

    return (user + tsys) ;
}

double WallTimer ( void )
{
    struct timeval tv ;
    double secs, usecs ;

    gettimeofday (&tv, NULL) ;
    secs = tv.tv_sec ;
    usecs = tv.tv_usec ;
    return (secs + usecs*1.e-6) ;
}
